import torch
import torchvision as tv

samples0, samples1 = 60000, 10000
classes = 10

source0 = tv.datasets.MNIST("../MNIST", train = True, download = True)
source1 = tv.datasets.MNIST("../MNIST", train = False, download = True)
DATA0 = source0.data.unsqueeze(1).float().cuda()
DATA1 = source1.data.unsqueeze(1).float().cuda()
TARGET0 = source0.targets.cuda()
TARGET1 = source1.targets.cuda()

FILTER1 = torch.zeros(8, 1, 5, 5, requires_grad = True, device = "cuda")
BIAS = torch.zeros(1, classes, requires_grad = True, device = "cuda")
WEIGHT = torch.zeros(8 * 24 * 24, classes, requires_grad = True, device = "cuda")
torch.nn.init.xavier_normal_(FILTER1)
variables = [BIAS, WEIGHT, FILTER1]

def model(SIGNAL):                                       # (batch, 1, 28, 28)
    SIGNAL = torch.nn.functional.conv2d(SIGNAL, FILTER1) # (batch, 8, 24, 24)
    SIGNAL = torch.nn.functional.relu(SIGNAL)
    SIGNAL = SIGNAL.flatten(1)                           # (batch, 8 * 24 * 24)
    SIGNAL = BIAS + SIGNAL @ WEIGHT                      # (batch, 10)
    return SIGNAL

batch = 1000
optimizer = torch.optim.Adam(variables, lr = 0.0001)
for epoch in range(100):
    LOSS0 = torch.zeros((), device = "cuda")
    ACCURACY0 = torch.zeros((), device = "cuda")
    count0 = 0
    for index in range(0, samples0, batch):
        optimizer.zero_grad()
        DATA = DATA0[index : index + batch]
        TARGET = TARGET0[index : index + batch]
        count = TARGET.size(0)
        ACTIVATION = model(DATA)
        LOSS = torch.nn.functional.cross_entropy(ACTIVATION, TARGET)
        LOSS0 += LOSS * count
        VALUE = torch.argmax(ACTIVATION, 1)
        ACCURACY0 += torch.sum(VALUE == TARGET)
        count0 += count
        LOSS.backward()
        optimizer.step()
    LOSS0 /= count0
    ACCURACY0 /= count0
    with torch.no_grad():
        LOSS1 = torch.zeros((), device = "cuda")
        ACCURACY1 = torch.zeros((), device = "cuda")
        count1 = 0
        for index in range(0, samples1, batch):
            DATA = DATA1[index : index + batch]
            TARGET = TARGET1[index : index + batch]
            ACTIVATION = model(DATA)
            LOSS1 += torch.nn.functional.cross_entropy(ACTIVATION, TARGET, reduction = "sum")
            VALUE = torch.argmax(ACTIVATION, 1)
            ACCURACY1 += torch.sum(VALUE == TARGET)
            count1 += TARGET.size(0)
        LOSS1 /= count1
        ACCURACY1 /= count1
    print("%5d %12.3f %4.3f %12.3f %4.3f" % \
          (epoch, LOSS0, ACCURACY0, LOSS1, ACCURACY1), flush = True)

#accuracy: train: 1000 test: 980
